import pandas as pd
import pytest
import numpy as np

from nempy import markets
from nempy.help_functions import helper_functions as hf
from nempy.historical_inputs import mms_db as hi, demand


class SpotMarketBuilder:
    def __init__(self, unit_inputs, interconnector_inputs, constraint_inputs, demand_inputs):

        self.unit_inputs = unit_inputs
        self.interconnector_inputs = interconnector_inputs
        self.constraint_inputs = constraint_inputs
        self.regional_demand_inputs = demand_inputs

        unit_info = self.unit_inputs.get_unit_info()
        self.market = markets.SpotMarket(market_regions=['QLD1', 'NSW1', 'VIC1', 'SA1', 'TAS1'], unit_info=unit_info)
        self.market.solver_name = 'CBC'

    def set_solver(self, solver_name):
        self.market.solver_name = solver_name

    def add_unit_bids_to_market(self):
        volume_bids, price_bids = self.unit_inputs.get_processed_bids()
        self.market.set_unit_volume_bids(volume_bids)
        self.market.set_unit_price_bids(price_bids)

    def set_unit_limit_constraints(self):
        unit_bid_limit = self.unit_inputs.get_unit_bid_availability()
        self.market.set_unit_bid_capacity_constraints(unit_bid_limit)
        cost = self.constraint_inputs.get_constraint_violation_prices()['unit_capacity']
        self.market.make_constraints_elastic('unit_bid_capacity', violation_cost=cost)
        unit_uigf_limit = self.unit_inputs.get_unit_uigf_limits()
        self.market.set_unconstrained_intermitent_generation_forecast_constraint(unit_uigf_limit)
        cost = self.constraint_inputs.get_constraint_violation_prices()['uigf']
        self.market.make_constraints_elastic('uigf_capacity', violation_cost=cost)

    def set_ramp_rate_limits(self):
        ramp_rates = self.unit_inputs.get_ramp_rates_used_for_energy_dispatch()
        self.market.set_unit_ramp_up_constraints(
            ramp_rates.loc[:, ['unit', 'initial_output', 'ramp_up_rate']])
        self.market.set_unit_ramp_down_constraints(
            ramp_rates.loc[:, ['unit', 'initial_output', 'ramp_down_rate']])
        cost = self.constraint_inputs.get_constraint_violation_prices()['ramp_rate']
        self.market.make_constraints_elastic('ramp_up', violation_cost=cost)
        self.market.make_constraints_elastic('ramp_down', violation_cost=cost)

    def set_fast_start_constraints(self):
        self.market.dispatch()
        dispatch = self.market.get_unit_dispatch()
        fast_start_profiles = self.unit_inputs.get_fast_start_profiles_for_dispatch(dispatch)
        self.market.set_fast_start_constraints(fast_start_profiles)
        if 'fast_start' in self.market._constraints_rhs_and_type:
            cost = self.constraint_inputs.get_constraint_violation_prices()['fast_start']
            self.market.make_constraints_elastic('fast_start', cost)

    def set_unit_fcas_constraints(self):
        self.unit_inputs.add_fcas_trapezium_constraints()

        cost = self.constraint_inputs.get_constraint_violation_prices()['fcas_max_avail']
        fcas_availability = self.unit_inputs.get_fcas_max_availability()
        self.market.set_fcas_max_availability(fcas_availability)
        self.market.make_constraints_elastic('fcas_max_availability', cost)

        cost = self.constraint_inputs.get_constraint_violation_prices()['fcas_profile']

        regulation_trapeziums = self.unit_inputs.get_fcas_regulation_trapeziums()
        self.market.set_energy_and_regulation_capacity_constraints(regulation_trapeziums)
        self.market.make_constraints_elastic('energy_and_regulation_capacity', cost)

        scada_ramp_down_rates = self.unit_inputs.get_scada_ramp_down_rates_of_lower_reg_units()
        self.market.set_joint_ramping_constraints_lower_reg(scada_ramp_down_rates)
        self.market.make_constraints_elastic('joint_ramping_lower_reg', cost)

        scada_ramp_up_rates = self.unit_inputs.get_scada_ramp_up_rates_of_raise_reg_units()
        self.market.set_joint_ramping_constraints_raise_reg(scada_ramp_up_rates)
        self.market.make_constraints_elastic('joint_ramping_raise_reg', cost)

        contingency_trapeziums = self.unit_inputs.get_contingency_services()
        self.market.set_joint_capacity_constraints(contingency_trapeziums)
        self.market.make_constraints_elastic('joint_capacity', cost)

    def set_region_demand_constraints(self):
        regional_demand = self.regional_demand_inputs.get_operational_demand()
        self.market.set_demand_constraints(regional_demand)
        cost = self.constraint_inputs.get_constraint_violation_prices()['regional_demand']
        self.market.make_constraints_elastic('demand', cost)

    def add_interconnectors_to_market(self):
        interconnectors = self.interconnector_inputs.get_interconnector_definitions()
        loss_functions, interpolation_break_points = self.interconnector_inputs.get_interconnector_loss_model()
        self.market.set_interconnectors(interconnectors)
        self.market.set_interconnector_losses(loss_functions, interpolation_break_points)

    def add_generic_constraints_with_fcas_requirements_interface(self):
        fcas_requirements = self.constraint_inputs.get_fcas_requirements()
        self.market.set_fcas_requirements_constraints(fcas_requirements)
        violation_costs = self.constraint_inputs.get_violation_costs()
        self.market.make_constraints_elastic('fcas', violation_cost=violation_costs)
        generic_rhs = self.constraint_inputs.get_rhs_and_type_excluding_regional_fcas_constraints()
        self.market.set_generic_constraints(generic_rhs)
        self.market.make_constraints_elastic('generic', violation_cost=violation_costs)
        unit_generic_lhs = self.constraint_inputs.get_unit_lhs()
        self.market.link_units_to_generic_constraints(unit_generic_lhs)
        interconnector_generic_lhs = self.constraint_inputs.get_interconnector_lhs()
        self.market.link_interconnectors_to_generic_constraints(interconnector_generic_lhs)

    def add_generic_constraints(self):
        violation_costs = self.constraint_inputs.get_violation_costs()
        generic_rhs = self.constraint_inputs.get_rhs_and_type()
        self.market.set_generic_constraints(generic_rhs)
        self.market.make_constraints_elastic('generic', violation_cost=violation_costs)
        unit_generic_lhs = self.constraint_inputs.get_unit_lhs()
        self.market.link_units_to_generic_constraints(unit_generic_lhs)
        interconnector_generic_lhs = self.constraint_inputs.get_interconnector_lhs()
        self.market.link_interconnectors_to_generic_constraints(interconnector_generic_lhs)
        regions_generic_lhs = self.constraint_inputs.get_region_lhs()
        self.market.link_regions_to_generic_constraints(regions_generic_lhs)

    def dispatch(self, calc_prices=True):
        if self.constraint_inputs.is_over_constrained_dispatch_rerun():
            self.market.dispatch(allow_over_constrained_dispatch_re_run=True,
                                 energy_market_floor_price=-1000.0, energy_market_ceiling_price=14500.0,
                                 fcas_market_ceiling_price=1000.0)
        else:
            self.market.dispatch(allow_over_constrained_dispatch_re_run=False)

    def get_market_object(self):
        return self.market


class MarketOverrider:
    def __init__(self, market, mms_db, interval):
        self.services = ['TOTALCLEARED', 'LOWER5MIN', 'LOWER60SEC', 'LOWER6SEC', 'RAISE5MIN', 'RAISE60SEC', 'RAISE6SEC',
                         'LOWERREG', 'RAISEREG']

        self.service_name_mapping = {'TOTALCLEARED': 'energy', 'RAISEREG': 'raise_reg', 'LOWERREG': 'lower_reg',
                                     'RAISE6SEC': 'raise_6s', 'RAISE60SEC': 'raise_60s', 'RAISE5MIN': 'raise_5min',
                                     'LOWER6SEC': 'lower_6s', 'LOWER60SEC': 'lower_60s', 'LOWER5MIN': 'lower_5min',
                                     'ENERGY': 'energy'}

        self.inputs_manager = mms_db
        self.interval = interval

        self.market = market

    def set_unit_dispatch_to_historical_values(self, wiggle_room=0.001):
        DISPATCHLOAD = self.inputs_manager.DISPATCHLOAD.get_data(self.interval)

        bounds = DISPATCHLOAD.loc[:, ['DUID'] + self.services]
        bounds.columns = ['unit'] + self.services

        bounds = hf.stack_columns(bounds, cols_to_keep=['unit'], cols_to_stack=self.services, type_name='service',
                                  value_name='dispatched')

        bounds['service'] = bounds['service'].apply(lambda x: self.service_name_mapping[x])

        decision_variables = self.market._decision_variables['bids'].copy()

        decision_variables = pd.merge(decision_variables, bounds, on=['unit', 'service'])

        decision_variables_first_bid = decision_variables.groupby(['unit', 'service'], as_index=False).first()

        def last_bids(df):
            return df.iloc[1:]

        decision_variables_remaining_bids = \
            decision_variables.groupby(['unit', 'service'], as_index=False).apply(last_bids)

        decision_variables_first_bid['lower_bound'] = decision_variables_first_bid['dispatched'] - wiggle_room
        decision_variables_first_bid['upper_bound'] = decision_variables_first_bid['dispatched'] + wiggle_room
        decision_variables_first_bid['lower_bound'] = np.where(decision_variables_first_bid['lower_bound'] < 0.0, 0.0,
                                                               decision_variables_first_bid['lower_bound'])
        decision_variables_first_bid['upper_bound'] = np.where(decision_variables_first_bid['upper_bound'] < 0.0, 0.0,
                                                               decision_variables_first_bid['upper_bound'])
        decision_variables_remaining_bids['lower_bound'] = 0.0
        decision_variables_remaining_bids['upper_bound'] = 0.0

        decision_variables = pd.concat([decision_variables_first_bid, decision_variables_remaining_bids])

        self.market._decision_variables['bids'] = decision_variables

    def set_interconnector_flow_to_historical_values(self, wiggle_room=0.1):
        # Historical interconnector dispatch
        DISPATCHINTERCONNECTORRES = self.inputs_manager.DISPATCHINTERCONNECTORRES.get_data(self.interval)
        interconnector_flow = DISPATCHINTERCONNECTORRES.loc[:, ['INTERCONNECTORID', 'MWFLOW']]
        interconnector_flow.columns = ['interconnector', 'flow']
        interconnector_flow['link'] = interconnector_flow['interconnector']
        interconnector_flow['link'] = np.where(interconnector_flow['interconnector'] == 'T-V-MNSP1',
                                               np.where(interconnector_flow['flow'] >= 0.0, 'BLNKTAS', 'BLNKVIC'),
                                               interconnector_flow['link'])

        flow_variables = self.market._decision_variables['interconnectors']
        flow_variables = pd.merge(flow_variables, interconnector_flow, 'left', on=['interconnector', 'link'])
        flow_variables = flow_variables.fillna(0.0)
        flow_variables['flow'] = np.where(flow_variables['link'] != flow_variables['interconnector'],
                                          flow_variables['flow'].abs(), flow_variables['flow'])
        flow_variables['lower_bound'] = flow_variables['flow'] - wiggle_room
        flow_variables['upper_bound'] = flow_variables['flow'] + wiggle_room
        flow_variables = flow_variables.drop(['flow'], axis=1)
        self.market._decision_variables['interconnectors'] = flow_variables


class MarketChecker:
    def __init__(self, market, mms_db, xml_cache, interval, unit_inputs=None):
        self.services = ['TOTALCLEARED', 'LOWER5MIN', 'LOWER60SEC', 'LOWER6SEC', 'RAISE5MIN', 'RAISE60SEC', 'RAISE6SEC',
                         'LOWERREG', 'RAISEREG']

        self.service_name_mapping = {'TOTALCLEARED': 'energy', 'RAISEREG': 'raise_reg', 'LOWERREG': 'lower_reg',
                                     'RAISE6SEC': 'raise_6s', 'RAISE60SEC': 'raise_60s', 'RAISE5MIN': 'raise_5min',
                                     'LOWER6SEC': 'lower_6s', 'LOWER60SEC': 'lower_60s', 'LOWER5MIN': 'lower_5min',
                                     'ENERGY': 'energy'}

        self.inputs_manager = mms_db
        self.xml = xml_cache
        self.unit_inputs = unit_inputs
        self.interval = interval

        self.market = market

    def all_dispatch_units_and_services_have_decision_variables(self, wiggle_room=0.001):
        DISPATCHLOAD = self.inputs_manager.DISPATCHLOAD.get_data(self.interval)

        bounds = DISPATCHLOAD.loc[:, ['DUID'] + self.services]
        bounds.columns = ['unit'] + self.services

        bounds = hf.stack_columns(bounds, cols_to_keep=['unit'], cols_to_stack=self.services, type_name='service',
                                  value_name='dispatched')

        bounds['service'] = bounds['service'].apply(lambda x: self.service_name_mapping[x])

        bounds = bounds[bounds['dispatched'] > 0.001]

        decision_variables = self.market._decision_variables['bids'].copy()

        decision_variables = decision_variables.groupby(['unit', 'service'], as_index=False).first()

        decision_variables = pd.merge(bounds, decision_variables, how='left', on=['unit', 'service'])

        decision_variables['not_missing'] = ~decision_variables['variable_id'].isna()

        decision_variables = decision_variables.sort_values('not_missing')

        return decision_variables['not_missing'].all()

    def is_regional_demand_meet(self, tolerance=0.5):
        DISPATCHREGIONSUM = self.inputs_manager.DISPATCHREGIONSUM.get_data(self.interval)
        regional_demand = demand._format_regional_demand(DISPATCHREGIONSUM)
        region_summary = self.market.get_region_dispatch_summary()
        region_summary = pd.merge(region_summary, regional_demand, on='region')
        region_summary['calc_demand'] = region_summary['dispatch'] + region_summary['inflow'] \
            - region_summary['interconnector_losses'] - region_summary['transmission_losses']
        region_summary['diff'] = region_summary['calc_demand'] - region_summary['demand']
        region_summary['no_error'] = region_summary['diff'].abs() < tolerance
        return region_summary['no_error'].all()

    def is_generic_constraint_slack_correct(self):

        def calc_slack(rhs, lhs, type):
            if type == '<=':
                slack = rhs - lhs
            elif type == '>=':
                slack = lhs - rhs
            else:
                slack = 0.0
            if slack < 0.0:
                slack = 0.0
            return slack

        DISPATCHCONSTRAINT = self.inputs_manager.DISPATCHCONSTRAINT.get_data(self.interval)
        generic_cons_slack = self.market._constraints_rhs_and_type['generic']
        generic_cons_slack['slack'] = generic_cons_slack['slack'].abs()
        generic_cons_slack = pd.merge(generic_cons_slack, DISPATCHCONSTRAINT, left_on='set',
                                      right_on='CONSTRAINTID')
        generic_cons_slack['aemo_slack'] = (generic_cons_slack['RHS'] - generic_cons_slack['LHS'])
        generic_cons_slack['aemo_slack'] = \
            generic_cons_slack.apply(lambda x: calc_slack(x['RHS'], x['LHS'], x['type']), axis=1)
        generic_cons_slack['comp'] = (generic_cons_slack['aemo_slack'] - generic_cons_slack['slack']).abs()
        generic_cons_slack['no_error'] = generic_cons_slack['comp'] < 0.9
        return generic_cons_slack['no_error'].all()

    def is_fcas_constraint_slack_correct(self):

        def calc_slack(rhs, lhs, type):
            if type == '<=':
                slack = rhs - lhs
            elif type == '>=':
                slack = lhs - rhs
            else:
                slack = 0.0
            if slack < 0.0:
                slack = 0.0
            return slack

        DISPATCHCONSTRAINT = self.inputs_manager.DISPATCHCONSTRAINT.get_data(self.interval)
        generic_cons_slack = self.market._market_constraints_rhs_and_type['fcas']
        generic_cons_slack['slack'] = generic_cons_slack['slack'].abs()
        generic_cons_slack = pd.merge(generic_cons_slack, DISPATCHCONSTRAINT, left_on='set',
                                      right_on='CONSTRAINTID')
        generic_cons_slack['aemo_slack'] = (generic_cons_slack['RHS'] - generic_cons_slack['LHS'])
        generic_cons_slack['aemo_slack'] = \
            generic_cons_slack.apply(lambda x: calc_slack(x['RHS'], x['LHS'], x['type']), axis=1)
        generic_cons_slack['comp'] = (generic_cons_slack['aemo_slack'] - generic_cons_slack['slack']).abs()
        generic_cons_slack['no_error'] = generic_cons_slack['comp'] < 0.9
        return generic_cons_slack['no_error'].all()

    def all_constraints_presenet(self):
        DISPATCHCONSTRAINT = list(self.inputs_manager.DISPATCHCONSTRAINT.get_data(self.interval)['CONSTRAINTID'])
        fcas = list(self.market._market_constraints_rhs_and_type['fcas']['set'])
        generic = list(self.market._constraints_rhs_and_type['generic']['set'])
        generic = generic + fcas
        return set(DISPATCHCONSTRAINT) < set(generic + [1])

    def get_price_comparison(self):
        energy_prices = self.market.get_energy_prices()
        energy_prices['time'] = self.interval
        energy_prices['service'] = 'energy'
        fcas_prices = self.market.get_fcas_prices()
        fcas_prices['time'] = self.interval
        prices = pd.concat([energy_prices, fcas_prices])

        price_to_service = {'ROP': 'energy', 'RAISE6SECROP': 'raise_6s', 'RAISE60SECROP': 'raise_60s',
                            'RAISE5MINROP': 'raise_5min', 'RAISEREGROP': 'raise_reg', 'LOWER6SECROP': 'lower_6s',
                            'LOWER60SECROP': 'lower_60s', 'LOWER5MINROP': 'lower_5min', 'LOWERREGROP': 'lower_reg'}
        price_columns = list(price_to_service.keys())
        historical_prices = self.inputs_manager.DISPATCHPRICE.get_data(self.interval)
        historical_prices = hf.stack_columns(historical_prices, cols_to_keep=['SETTLEMENTDATE', 'REGIONID'],
                                             cols_to_stack=price_columns, type_name='service',
                                             value_name='RRP')
        historical_prices['service'] = historical_prices['service'].apply(lambda x: price_to_service[x])
        historical_prices = historical_prices.loc[:, ['SETTLEMENTDATE', 'REGIONID', 'service', 'RRP']]
        historical_prices.columns = ['time', 'region', 'service', 'hist_price']
        prices = pd.merge(prices, historical_prices, on=['time', 'region', 'service'])
        prices['error'] = prices['price'] - prices['hist_price']
        return prices

    def get_dispatch_comparison(self):
        DISPATCHLOAD = self.inputs_manager.DISPATCHLOAD.get_data(self.interval)
        bounds = DISPATCHLOAD.loc[:, ['DUID'] + self.services]
        bounds.columns = ['unit'] + self.services
        bounds = hf.stack_columns(bounds, cols_to_keep=['unit'], cols_to_stack=self.services, type_name='service',
                                  value_name='dispatched')
        bounds['service'] = bounds['service'].apply(lambda x: self.service_name_mapping[x])

        nempy_dispatch = self.market.get_unit_dispatch()
        comp = pd.merge(bounds, nempy_dispatch, 'inner', on=['unit', 'service'])
        comp['diff'] = comp['dispatch'] - comp['dispatched']
        comp = pd.merge(comp, self.market._unit_info.loc[:, ['unit', 'dispatch_type']], on='unit')
        comp['diff'] = np.where((comp['dispatch_type'] == 'load') & (comp['service'] == 'energy'), comp['diff'] * -1,
                                comp['diff'])
        return comp

    def do_fcas_availabilities_match_historical(self):
        DISPATCHLOAD = self.inputs_manager.DISPATCHLOAD.get_data(self.interval)
        availabilities = ['RAISE6SECACTUALAVAILABILITY', 'RAISE60SECACTUALAVAILABILITY',
                          'RAISE5MINACTUALAVAILABILITY', 'RAISEREGACTUALAVAILABILITY',
                          'LOWER6SECACTUALAVAILABILITY', 'LOWER60SECACTUALAVAILABILITY',
                          'LOWER5MINACTUALAVAILABILITY', 'LOWERREGACTUALAVAILABILITY']

        availabilities_mapping = {'RAISEREGACTUALAVAILABILITY': 'raise_reg',
                                  'LOWERREGACTUALAVAILABILITY': 'lower_reg',
                                  'RAISE6SECACTUALAVAILABILITY': 'raise_6s',
                                  'RAISE60SECACTUALAVAILABILITY': 'raise_60s',
                                  'RAISE5MINACTUALAVAILABILITY': 'raise_5min',
                                  'LOWER6SECACTUALAVAILABILITY': 'lower_6s',
                                  'LOWER60SECACTUALAVAILABILITY': 'lower_60s',
                                  'LOWER5MINACTUALAVAILABILITY': 'lower_5min'}

        bounds = DISPATCHLOAD.loc[:, ['DUID'] + availabilities]
        bounds.columns = ['unit'] + availabilities

        availabilities = hf.stack_columns(bounds, cols_to_keep=['unit'], cols_to_stack=availabilities,
                                          type_name='service', value_name='availability')

        bounds = DISPATCHLOAD.loc[:, ['DUID'] + self.services]
        bounds.columns = ['unit'] + self.services

        bounds = hf.stack_columns(bounds, cols_to_keep=['unit'], cols_to_stack=self.services, type_name='service',
                                  value_name='dispatched')

        bounds['service'] = bounds['service'].apply(lambda x: self.service_name_mapping[x])

        availabilities['service'] = availabilities['service'].apply(lambda x: availabilities_mapping[x])

        #availabilities = pd.merge(availabilities, bounds, on=['unit', 'service'])

        #availabilities = availabilities[~(availabilities['dispatched'] - 0.001 > availabilities['availability'])]

        output = self.market.get_fcas_availability()
        output.columns = ['unit', 'service', 'availability_measured']

        availabilities = pd.merge(availabilities, output, 'left', on=['unit', 'service'])

        availabilities['availability_measured'] = availabilities['availability_measured'].fillna(0)

        availabilities['error'] = availabilities['availability_measured'] - availabilities['availability']

        availabilities['match'] = availabilities['error'].abs() < 0.1
        availabilities = availabilities.sort_values('match')

        return availabilities

    def measured_violation_equals_historical_violation(self, historical_name, nempy_constraints):
        measured = 0.0
        for name in nempy_constraints:
            measured += self.market.get_elastic_constraints_violation_degree(name)
        historical = self.xml.get_violations()[historical_name]
        return measured == pytest.approx(historical, abs=0.1)



class RawInputsLoader:
    """Provides single interface for accessing raw historical inputs.

    Examples
    --------

    >>> import sqlite3

    >>> from nempy.historical_inputs import mms_db
    >>> from nempy.historical_inputs import xml_cache

    For the RawInputsLoader to work we need to construct a database and inputs cache for it to load inputs from and then
    pass the interfaces to these to the inputs loader.

    >>> con = sqlite3.connect('market_management_system.db')
    >>> mms_db_manager = mms_db.DBManager(connection=con)
    >>> xml_cache_manager = xml_cache.XMLCacheManager('test_nemde_cache')

    In this example the database and cache have already been populated so the input loader can be created straight
    away.

    >>> inputs_loader = RawInputsLoader(xml_cache_manager, mms_db_manager)

    Then we set the dispatch interval that we want to load inputs from.

    >>> inputs_loader.set_interval('2019/01/01 00:00:00')

    And then we can load some inputs.

    >>> inputs_loader.get_unit_volume_bids()
             DUID     BIDTYPE  MAXAVAIL  ENABLEMENTMIN  ENABLEMENTMAX  LOWBREAKPOINT  HIGHBREAKPOINT  BANDAVAIL1  BANDAVAIL2  BANDAVAIL3  BANDAVAIL4  BANDAVAIL5  BANDAVAIL6  BANDAVAIL7  BANDAVAIL8  BANDAVAIL9  BANDAVAIL10  RAMPDOWNRATE  RAMPUPRATE
    0      AGLHAL      ENERGY     173.0            0.0            0.0            0.0             0.0         0.0         0.0         0.0         0.0         0.0         0.0        60.0         0.0         0.0        160.0         720.0       720.0
    1      AGLSOM      ENERGY     160.0            0.0            0.0            0.0             0.0         0.0         0.0         0.0         0.0         0.0         0.0         0.0         0.0         0.0        170.0         480.0       480.0
    2     ANGAST1      ENERGY      43.0            0.0            0.0            0.0             0.0         0.0         0.0         0.0         0.0         0.0        50.0         0.0         0.0         0.0         50.0         840.0       840.0
    3       APD01   LOWER5MIN       0.0            0.0            0.0            0.0             0.0         0.0         0.0         0.0         0.0         0.0         0.0         0.0         0.0         0.0        300.0           0.0         0.0
    4       APD01  LOWER60SEC       0.0            0.0            0.0            0.0             0.0         0.0         0.0         0.0         0.0         0.0         0.0         0.0         0.0         0.0        300.0           0.0         0.0
    ...       ...         ...       ...            ...            ...            ...             ...         ...         ...         ...         ...         ...         ...         ...         ...         ...          ...           ...         ...
    1021    YWPS4   LOWER6SEC      25.0          250.0          385.0          275.0           385.0        15.0        10.0         0.0         0.0         0.0         0.0         0.0         0.0         0.0          0.0           0.0         0.0
    1022    YWPS4   RAISE5MIN       0.0          250.0          390.0          250.0           380.0         0.0         0.0         0.0         0.0         5.0         0.0         0.0         5.0         0.0         10.0           0.0         0.0
    1023    YWPS4    RAISEREG      15.0          250.0          385.0          250.0           370.0         0.0         0.0         0.0         0.0         0.0         0.0         5.0        10.0         0.0          5.0           0.0         0.0
    1024    YWPS4  RAISE60SEC      10.0          220.0          400.0          220.0           390.0         0.0         0.0         0.0         0.0         0.0         5.0         5.0         0.0         0.0         10.0           0.0         0.0
    1025    YWPS4   RAISE6SEC      15.0          220.0          405.0          220.0           390.0         0.0         0.0         0.0        10.0         5.0         0.0         0.0         0.0         0.0         10.0           0.0         0.0
    <BLANKLINE>
    [1026 rows x 19 columns]

    """
    def __init__(self, nemde_xml_cache_manager, market_management_system_database):
        self.xml = nemde_xml_cache_manager
        self.mms_db = market_management_system_database
        self.interval = None

    def set_interval(self, interval):
        """Set the interval to load inputs for.

        Examples
        --------

        For an example see the :func:`class level documentation <nempy.historical_inputs.loaders.RawInputsLoader>`


        Parameters
        ----------
        interval : str
            In the format '%Y/%m/%d %H:%M:%S'

        """
        self.interval = interval
        self.xml.load_interval(interval)

    def get_unit_initial_conditions(self):
        """Direct interface to :func:`nempy.historical_inputs.xml_cache.XMLCacheManager.get_unit_initial_conditions <nempy.historical_inputs.xml_cache.XMLCacheManager.get_unit_initial_conditions>`
        """
        return self.xml.get_unit_initial_conditions()

    def get_unit_volume_bids(self):
        """Direct interface to :func:`nempy.historical_inputs.xml_cache.XMLCacheManager.get_unit_volume_bids <nempy.historical_inputs.xml_cache.XMLCacheManager.get_unit_volume_bids>`
        """
        return self.xml.get_unit_volume_bids()

    def get_unit_price_bids(self):
        """Direct interface to :attr:`nempy.historical_inputs.mms_db.DBManager.BIDDAYOFFER_D.get_data <nempy.historical_inputs.mms_db.DBManager.BIDDAYOFFER_D>`
        """
        # If you change this to source data from MMS then you need to also change units.UnitData.get_processed_bids
        # to not undo scaling by loss factors.
        return self.xml.get_unit_price_bids()

    def get_unit_details(self):
        """Direct interface to :attr:`nempy.historical_inputs.mms_db.DBManager.DUDETAILSUMMARY.get_data <nempy.historical_inputs.mms_db.DBManager.DUDETAILSUMMARY>`
        """
        return self.mms_db.DUDETAILSUMMARY.get_data(self.interval)

    def get_agc_enablement_limits(self):
        """Direct interface to :attr:`nempy.historical_inputs.mms_db.DBManager.DISPATCHLOAD.get_data <nempy.historical_inputs.mms_db.DBManager.DISPATCHLOAD>`
        """
        return self.mms_db.DISPATCHLOAD.get_data(self.interval)

    def get_UIGF_values(self):
        """Direct interface to :func:`nempy.historical_inputs.xml_cache.XMLCacheManager.get_UIGF_values <nempy.historical_inputs.xml_cache.XMLCacheManager.get_UIGF_values>`
        """
        return self.xml.get_UIGF_values()

    def get_violations(self):
        """Direct interface to :func:`nempy.historical_inputs.xml_cache.XMLCacheManager.get_violations <nempy.historical_inputs.xml_cache.XMLCacheManager.get_violations>`
        """
        return self.xml.get_violations()

    def get_constraint_violation_prices(self):
        """Direct interface to :func:`nempy.historical_inputs.xml_cache.XMLCacheManager.get_constraint_violation_prices <nempy.historical_inputs.xml_cache.XMLCacheManager.get_constraint_violation_prices>`
        """
        return self.xml.get_constraint_violation_prices()

    def get_constraint_rhs(self):
        """Direct interface to :func:`nempy.historical_inputs.xml_cache.XMLCacheManager.get_constraint_rhs <nempy.historical_inputs.xml_cache.XMLCacheManager.get_constraint_rhs>`
        """
        return self.xml.get_constraint_rhs()

    def get_constraint_type(self):
        """Direct interface to :func:`nempy.historical_inputs.xml_cache.XMLCacheManager.get_constraint_type <nempy.historical_inputs.xml_cache.XMLCacheManager.get_constraint_type>`
        """
        return self.xml.get_constraint_type()

    def get_constraint_region_lhs(self):
        """Direct interface to :func:`nempy.historical_inputs.xml_cache.XMLCacheManager.get_constraint_region_lhs <nempy.historical_inputs.xml_cache.XMLCacheManager.get_constraint_region_lhs>`
        """
        return self.xml.get_constraint_region_lhs()

    def get_constraint_unit_lhs(self):
        """Direct interface to :func:`nempy.historical_inputs.xml_cache.XMLCacheManager.get_constraint_unit_lhs <nempy.historical_inputs.xml_cache.XMLCacheManager.get_constraint_unit_lhs>`
        """
        return self.xml.get_constraint_unit_lhs()

    def get_constraint_interconnector_lhs(self):
        """Direct interface to :func:`nempy.historical_inputs.xml_cache.XMLCacheManager.get_constraint_interconnector_lhs <nempy.historical_inputs.xml_cache.XMLCacheManager.get_constraint_interconnector_lhs>`
        """
        return self.xml.get_constraint_interconnector_lhs()

    def get_market_interconnectors(self):
        """Direct interface to :attr:`nempy.historical_inputs.mms_db.DBManager.MNSP_INTERCONNECTOR.get_data <nempy.historical_inputs.mms_db.DBManager.MNSP_INTERCONNECTOR>`
        """
        return self.mms_db.MNSP_INTERCONNECTOR.get_data(self.interval)

    def get_market_interconnector_link_bid_availability(self):
        """Direct interface to :func:`nempy.historical_inputs.xml_cache.XMLCacheManager.get_market_interconnector_link_bid_availability <nempy.historical_inputs.xml_cache.XMLCacheManager.get_market_interconnector_link_bid_availability>`
        """
        return self.xml.get_market_interconnector_link_bid_availability()

    def get_interconnector_constraint_parameters(self):
        """Direct interface to :attr:`nempy.historical_inputs.mms_db.DBManager.INTERCONNECTORCONSTRAINT.get_data <nempy.historical_inputs.mms_db.DBManager.INTERCONNECTORCONSTRAINT>`
        """
        return self.mms_db.INTERCONNECTORCONSTRAINT.get_data(self.interval)

    def get_interconnector_definitions(self):
        """Direct interface to :attr:`nempy.historical_inputs.mms_db.DBManager.INTERCONNECTOR.get_data <nempy.historical_inputs.mms_db.DBManager.INTERCONNECTOR>`
        """
        return self.mms_db.INTERCONNECTOR.get_data()

    def get_regional_loads(self):
        """Direct interface to :attr:`nempy.historical_inputs.mms_db.DBManager.DISPATCHREGIONSUM.get_data <nempy.historical_inputs.mms_db.DBManager.DISPATCHREGIONSUM>`
        """
        return self.mms_db.DISPATCHREGIONSUM.get_data(self.interval)

    def get_interconnector_loss_segments(self):
        """Direct interface to :attr:`nempy.historical_inputs.mms_db.DBManager.LOSSMODEL.get_data <nempy.historical_inputs.mms_db.DBManager.LOSSMODEL>`
        """
        return self.mms_db.LOSSMODEL.get_data(self.interval)

    def get_interconnector_loss_parameters(self):
        """Direct interface to :attr:`nempy.historical_inputs.mms_db.DBManager.LOSSFACTORMODEL.get_data <nempy.historical_inputs.mms_db.DBManager.LOSSFACTORMODEL>`
        """
        return self.mms_db.LOSSFACTORMODEL.get_data(self.interval)

    def get_unit_fast_start_parameters(self):
        """Direct interface to :func:`nempy.historical_inputs.xml_cache.XMLCacheManager.get_unit_fast_start_parameters <nempy.historical_inputs.xml_cache.XMLCacheManager.get_unit_fast_start_parameters>`
        """
        return self.xml.get_unit_fast_start_parameters()

    def is_over_constrained_dispatch_rerun(self):
        """Checks if the over constrained dispatch rerun process was used by AEMO to dispatch this interval.

        Examples
        --------

        >>> import sqlite3

        >>> from nempy.historical_inputs import mms_db
        >>> from nempy.historical_inputs import xml_cache

        For the RawInputsLoader to work we need to construct a database and inputs cache for it to load inputs from and then
        pass the interfaces to these to the inputs loader.

        >>> con = sqlite3.connect('market_management_system.db')
        >>> mms_db_manager = mms_db.DBManager(connection=con)
        >>> xml_cache_manager = xml_cache.XMLCacheManager('test_nemde_cache')

        In this example the database and cache have already been populated so the input loader can be created straight
        away.

        >>> inputs_loader = RawInputsLoader(xml_cache_manager, mms_db_manager)

        Then we set the dispatch interval that we want to load inputs from.

        >>> inputs_loader.set_interval('2019/01/01 00:00:00')

        And then we can load some inputs.

        >>> inputs_loader.is_over_constrained_dispatch_rerun()
        False


        Returns
        -------
        bool

        """
        return 'OCD' in self.xml.get_file_name()



import numpy as np
import pandas as pd
from nempy.help_functions import helper_functions as hf


def create(definitions, next_variable_id):
    """Create decision variables, and their mapping to constraints. For modeling interconnector flows. As DataFrames.

    Examples
    --------
    Definitions for two interconnectors, one called A, that nominal flows from region X to region Y, note A can flow in
    both directions because of the way max and min are defined. The interconnector B nominal flows from Y to Z, but can
    only flow in the forward direction.

    >>> pd.options.display.width = None

    >>> inter_definitions = pd.DataFrame({
    ...   'interconnector': ['A', 'B'],
    ...   'link': ['A', 'B'],
    ...   'from_region': ['X', 'Y'],
    ...   'to_region': ['Y', 'Z'],
    ...   'max': [100.0, 400.0],
    ...   'min': [-100.0, 50.0],
    ...   'generic_constraint_factor': [1, 1],
    ...   'from_region_loss_factor': [0.9, 1.0],
    ...   'to_region_loss_factor': [1.0, 1.1]})

    >>> print(inter_definitions)
      interconnector link from_region to_region    max    min  generic_constraint_factor  from_region_loss_factor  to_region_loss_factor
    0              A    A           X         Y  100.0 -100.0                          1                      0.9                    1.0
    1              B    B           Y         Z  400.0   50.0                          1                      1.0                    1.1

    Start creating new variable ids from 0.

    >>> next_variable_id = 0

    Run the function and print results.

    >>> decision_variables, constraint_map = create(inter_definitions, next_variable_id)

    >>> print(decision_variables)
      interconnector link  variable_id  lower_bound  upper_bound        type  generic_constraint_factor
    0              A    A            0       -100.0        100.0  continuous                          1
    1              B    B            1         50.0        400.0  continuous                          1

    >>> print(constraint_map)
       variable_id interconnector link region service  coefficient
    0            0              A    A      Y  energy          1.0
    1            1              B    B      Z  energy          1.1
    2            0              A    A      X  energy         -0.9
    3            1              B    B      Y  energy         -1.0

    """

    # Create a variable_id for each interconnector.
    decision_variables = hf.save_index(definitions, 'variable_id', next_variable_id)

    # Create two entries in the constraint_map for each interconnector. This means the variable will be mapped to the
    # demand constraint of both connected regions.
    constraint_map = hf.stack_columns(decision_variables, ['variable_id', 'interconnector', 'link', 'max', 'min'],
                                      ['to_region', 'from_region'], 'direction', 'region')
    loss_factors = hf.stack_columns(decision_variables, ['variable_id'],
                                    ['from_region_loss_factor', 'to_region_loss_factor'], 'direction', 'loss_factor')
    loss_factors['direction'] = loss_factors['direction'].apply(lambda x: x.replace('_loss_factor', ''))
    constraint_map = pd.merge(constraint_map, loss_factors, on=['variable_id', 'direction'])

    # Define decision variable attributes.
    decision_variables['type'] = 'continuous'
    decision_variables = decision_variables.loc[:, ['interconnector', 'link', 'variable_id', 'min', 'max', 'type',
                                                    'generic_constraint_factor']]
    decision_variables.columns = ['interconnector', 'link',  'variable_id', 'lower_bound', 'upper_bound', 'type',
                                  'generic_constraint_factor']

    # Set positive coefficient for the to_region so the interconnector flowing in the nominal direction helps meet the
    # to_region demand constraint. Negative for the from_region, same logic.
    constraint_map['coefficient'] = np.where(constraint_map['direction'] == 'to_region',
                                             1.0 * constraint_map['loss_factor'],
                                             -1.0 * constraint_map['loss_factor'])
    constraint_map['service'] = 'energy'
    constraint_map = constraint_map.loc[:, ['variable_id', 'interconnector', 'link',
                                            'region', 'service', 'coefficient']]

    return decision_variables, constraint_map


def link_inter_loss_to_interpolation_weights(weight_variables, loss_variables, loss_functions, next_constraint_id):
    """
    Examples
    --------

    Setup function inputs

    >>> loss_variables = pd.DataFrame({
    ...   'interconnector': ['I'],
    ...   'link': ['I'],
    ...   'variable_id': [0]})

    >>> weight_variables = pd.DataFrame({
    ...   'interconnector': ['I', 'I', 'I'],
    ...   'link': ['I', 'I', 'I'],
    ...   'variable_id': [1, 2, 3],
    ...   'break_point': [-100.0, 0, 100.0]})

    Loss functions can arbitrary, they just need to take the flow as input and return losses as an output.

    >>> def constant_losses(flow):
    ...     return abs(flow) * 0.05

    The loss function get assigned to an interconnector by its row in the loss functions DataFrame.

    >>> loss_functions = pd.DataFrame({
    ...    'interconnector': ['I'],
    ...    'link': ['I'],
    ...    'from_region_loss_share': [0.5],
    ...    'loss_function': [constant_losses]})

    >>> next_constraint_id = 0

    Create the constraints.

    >>> lhs, rhs = link_inter_loss_to_interpolation_weights(weight_variables, loss_variables, loss_functions,
    ...                                                     next_constraint_id)

    >>> print(lhs)
       variable_id  constraint_id  coefficient
    0            1              0          5.0
    1            2              0          0.0
    2            3              0          5.0

    >>> print(rhs)
      interconnector link  constraint_id type  rhs_variable_id
    0              I    I              0    =                0
    """

    # Create a constraint for each set of weight variables.
    constraint_ids = weight_variables.loc[:, ['interconnector', 'link']].drop_duplicates(['interconnector', 'link'])
    constraint_ids = hf.save_index(constraint_ids, 'constraint_id', next_constraint_id)

    # Map weight variables to their corresponding constraints.
    lhs = pd.merge(weight_variables.loc[:, ['interconnector', 'link', 'variable_id', 'break_point']],
                   constraint_ids, 'inner', on=['interconnector', 'link'])
    lhs = pd.merge(lhs, loss_functions.loc[:, ['interconnector', 'link', 'loss_function']], 'inner',
                   on=['interconnector', 'link'])

    # Evaluate the loss function at each break point to get the lhs coefficient.
    lhs['coefficient'] = lhs.apply(lambda x: x['loss_function'](x['break_point']), axis=1)
    lhs = lhs.loc[:, ['variable_id', 'constraint_id', 'coefficient']]

    # Get the loss variables that will be on the rhs of the constraints.
    rhs_variables = loss_variables.loc[:, ['interconnector', 'link', 'variable_id']]
    rhs_variables.columns = ['interconnector', 'link', 'rhs_variable_id']
    # Map the rhs variables to their constraints.
    rhs = pd.merge(constraint_ids, rhs_variables, 'inner', on=['interconnector', 'link'])
    rhs['type'] = '='
    rhs = rhs.loc[:, ['interconnector', 'link', 'constraint_id', 'type', 'rhs_variable_id']]
    return lhs, rhs


def link_weights_to_inter_flow(weight_variables, flow_variables, next_constraint_id):
    """
    Examples
    --------

    Setup function inputs

    >>> flow_variables = pd.DataFrame({
    ...   'interconnector': ['I'],
    ...   'link': ['I'],
    ...   'variable_id': [0]})

    >>> weight_variables = pd.DataFrame({
    ...   'interconnector': ['I', 'I', 'I'],
    ...   'link': ['I', 'I', 'I'],
    ...   'variable_id': [1, 2, 3],
    ...   'break_point': [-100.0, 0, 100.0]})

    >>> next_constraint_id = 0

    Create the constraints.

    >>> lhs, rhs = link_weights_to_inter_flow(weight_variables, flow_variables, next_constraint_id)

    >>> print(lhs)
       variable_id  constraint_id  coefficient
    0            1              0       -100.0
    1            2              0          0.0
    2            3              0        100.0

    >>> print(rhs)
      interconnector link  constraint_id type  rhs_variable_id
    0              I    I              0    =                0
    """

    # Create a constraint for each set of weight variables.
    constraint_ids = weight_variables.loc[:, ['interconnector', 'link']].drop_duplicates(['interconnector', 'link'])
    constraint_ids = hf.save_index(constraint_ids, 'constraint_id', next_constraint_id)

    # Map weight variables to their corresponding constraints.
    lhs = pd.merge(weight_variables.loc[:, ['interconnector', 'link', 'variable_id', 'break_point']],
                   constraint_ids, 'inner', on=['interconnector', 'link'])
    lhs['coefficient'] = lhs['break_point']
    lhs = lhs.loc[:, ['variable_id', 'constraint_id', 'coefficient']]

    # Get the interconnector variables that will be on the rhs of constraint.
    rhs_variables = flow_variables.loc[:, ['interconnector', 'link', 'variable_id']]
    rhs_variables.columns = ['interconnector', 'link', 'rhs_variable_id']
    # Map the rhs variables to their constraints.
    rhs = pd.merge(constraint_ids, rhs_variables, 'inner', on=['interconnector', 'link'])
    rhs['type'] = '='
    rhs = rhs.loc[:, ['interconnector', 'link', 'constraint_id', 'type', 'rhs_variable_id']]
    return lhs, rhs


def create_weights_must_sum_to_one(weight_variables, next_constraint_id):
    """Create the constraint to force weight variable to sum to one, need for interpolation to work.

    For one interconnector, if we had  three weight variables w1, w2, and w3, then the constraint would be of the form.

        w1 * 1.0 + w2 * 1.0 + w3 * 1.0 = 1.0

    Examples
    --------

    Setup function inputs

    >>> weight_variables = pd.DataFrame({
    ...   'interconnector': ['I', 'I', 'I'],
    ...   'link': ['I', 'I', 'I'],
    ...   'variable_id': [1, 2, 3],
    ...   'break_point': [-100.0, 0, 100.0]})

    >>> next_constraint_id = 0

    Create the constraints.

    >>> lhs, rhs = create_weights_must_sum_to_one(weight_variables, next_constraint_id)

    >>> print(lhs)
       variable_id  constraint_id  coefficient
    0            1              0          1.0
    1            2              0          1.0
    2            3              0          1.0

    >>> print(rhs)
      interconnector link  constraint_id type  rhs
    0              I    I              0    =  1.0

    """

    # Create a constraint for each set of weight variables.
    constraint_ids = weight_variables.loc[:, ['interconnector', 'link']].drop_duplicates(['interconnector', 'link'])
    constraint_ids = hf.save_index(constraint_ids, 'constraint_id', next_constraint_id)

    # Map weight variables to their corresponding constraints.
    lhs = pd.merge(weight_variables.loc[:, ['interconnector', 'link', 'variable_id']], constraint_ids,
                   'inner', on=['interconnector', 'link'])
    lhs['coefficient'] = 1.0
    lhs = lhs.loc[:, ['variable_id', 'constraint_id', 'coefficient']]

    # Create rhs details for each constraint.
    rhs = constraint_ids
    rhs['type'] = '='
    rhs['rhs'] = 1.0
    return lhs, rhs


def create_weights(break_points, next_variable_id):
    """Create interpolation weight variables for each breakpoint.

    Examples
    --------

    >>> break_points = pd.DataFrame({
    ...   'interconnector': ['I', 'I', 'I'],
    ...   'loss_segment': [1, 2, 3],
    ...   'break_point': [-100.0, 0.0, 100.0]})

    >>> next_variable_id = 0

    >>> weight_variables = create_weights(break_points, next_variable_id)

    >>> print(weight_variables.loc[:, ['interconnector', 'loss_segment', 'break_point', 'variable_id']])
      interconnector  loss_segment  break_point  variable_id
    0              I             1       -100.0            0
    1              I             2          0.0            1
    2              I             3        100.0            2

    >>> print(weight_variables.loc[:, ['variable_id', 'lower_bound', 'upper_bound', 'type']])
       variable_id  lower_bound  upper_bound        type
    0            0          0.0          1.0  continuous
    1            1          0.0          1.0  continuous
    2            2          0.0          1.0  continuous

    Parameters
    ----------
    break_points : pd.DataFrame
        ==============  ================================================================================
        Columns:        Description:
        interconnector  unique identifier of a interconnector (as `str`)
        loss_segment    unique identifier of a loss segment on an interconnector basis (as `np.float64`)
        break_points    the interconnector flow values to interpolate losses between (as `np.float64`)
        ==============  ================================================================================

    next_variable_id : int

    Returns
    -------
    weight_variables : pd.DataFrame

        ==============  ==============================================================================
        Columns:        Description:
        interconnector  unique identifier of a interconnector (as `str`)
        loss_segment    unique identifier of a loss segment on an interconnector basis (as `np.float64`)
        break_points    the interconnector flow values to interpolate losses between (as `np.int64`)
        variable_id     the id of the variable (as `np.int64`)
        lower_bound    the lower bound of the variable, is zero for weight variables (as `np.float64`)
        upper_bound    the upper bound of the variable, is one for weight variables (as `np.float64`)
        type           the type of variable, is continuous for bids  (as `str`)
        ==============  ==============================================================================
    """
    # Create a variable for each break point.
    weight_variables = hf.save_index(break_points, 'variable_id', next_variable_id)
    weight_variables['lower_bound'] = 0.0
    weight_variables['upper_bound'] = 1.0
    weight_variables['type'] = 'continuous'
    return weight_variables


def create_loss_variables(inter_variables, inter_constraint_map, loss_shares, next_variable_id):
    """
    Examples
    --------
    Setup function inputs

    >>> inter_variables = pd.DataFrame({
    ...   'interconnector': ['I'],
    ...   'link': ['i'],
    ...   'lower_bound': [-50.0],
    ...   'upper_bound': [100.0],
    ...   'type': ['continuous']})

    >>> inter_constraint_map = pd.DataFrame({
    ... 'interconnector': ['I', 'I'],
    ... 'link': ['i', 'i'],
    ... 'region': ['X', 'Y'],
    ... 'service': ['energy', 'energy'],
    ... 'coefficient': [1.0, -1.0]})

    >>> loss_shares = pd.DataFrame({
    ...    'interconnector': ['I'],
    ...    'link': ['i'],
    ...    'from_region_loss_share': [0.5],
    ...    'from_region': ['X']})

    >>> next_constraint_id = 0

    Create the constraints.

    >>> loss_variables, constraint_map = create_loss_variables(inter_variables, inter_constraint_map, loss_shares,
    ...                                                        next_constraint_id)

    >>> print(loss_variables)
      interconnector link  variable_id  lower_bound  upper_bound        type
    0              I    i            0       -100.0        100.0  continuous

    >>> print(constraint_map)
       variable_id region service  coefficient
    0            0      X  energy         -0.5
    1            0      Y  energy         -0.5
    """

    # Preserve the interconnector variable id for merging later.
    columns_for_loss_variables = inter_variables.loc[:, ['interconnector', 'link', 'lower_bound',
                                                         'upper_bound', 'type']]
    columns_for_loss_variables['upper_bound'] = \
        columns_for_loss_variables.loc[:, ['lower_bound', 'upper_bound']].abs().max(axis=1)
    columns_for_loss_variables['lower_bound'] = -1 * columns_for_loss_variables['upper_bound']

    inter_constraint_map = inter_constraint_map.loc[:, ['interconnector', 'link', 'region', 'service', 'coefficient']]

    # Create a variable id for loss variables
    loss_variables = hf.save_index(columns_for_loss_variables, 'variable_id', next_variable_id)
    loss_variables = pd.merge(loss_variables,
                              loss_shares.loc[:, ['interconnector', 'link', 'from_region_loss_share', 'from_region']],
                              on=['interconnector', 'link'])

    # Create the loss variable constraint map by combining the new variables and the flow variable constraint map.
    constraint_map = pd.merge(
        loss_variables.loc[:, ['variable_id', 'interconnector', 'link', 'from_region_loss_share', 'from_region']],
        inter_constraint_map, 'inner', on=['interconnector', 'link'])

    # Assign losses to regions according to the from_region_loss_share
    constraint_map['coefficient'] = np.where(constraint_map['from_region'] == constraint_map['region'],
                                             - 1 * constraint_map['from_region_loss_share'],
                                             - 1 * (1 - constraint_map['from_region_loss_share']))

    loss_variables = loss_variables.loc[:, ['interconnector', 'link', 'variable_id',
                                            'lower_bound', 'upper_bound', 'type']]
    constraint_map = constraint_map.loc[:, ['variable_id', 'region', 'service', 'coefficient']]
    return loss_variables, constraint_map
